@@ -1,37 +1,20 @@
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
-from __future__ import absolute_import
-
 import awkward as ak
 
 np = ak.nplike.NumpyMetadata.instance()
 
 
-class Reducer(object):
+class Reducer:
     needs_position = False
 
     @classmethod
     def return_dtype(cls, given_dtype):
-        if (
-            given_dtype == np.bool_
-            or given_dtype == np.int8
-            or given_dtype == np.int16
-            or given_dtype == np.int32
-        ):
-            if ak._util.win or ak._util.bits32:
-                return np.int32
-            else:
-                return np.int64
+        if given_dtype in (np.bool_, np.int8, np.int16, np.int32):
+            return np.int32 if ak._util.win or ak._util.bits32 else np.int64
 
-        if (
-            given_dtype == np.uint8
-            or given_dtype == np.uint16
-            or given_dtype == np.uint32
-        ):
-            if ak._util.win or ak._util.bits32:
-                return np.uint32
-            else:
-                return np.uint64
+        if given_dtype in (np.uint8, np.uint16, np.uint32):
+            return np.uint32 if ak._util.win or ak._util.bits32 else np.uint64
 
         if given_dtype == np.complex128:
             return np.float64
@@ -42,7 +25,7 @@ class Reducer(object):
 
     @classmethod
     def maybe_double_length(cls, type, length):
-        return 2 * length if type == np.complex128 or type == np.complex64 else length
+        return 2 * length if type in (np.complex128, np.complex64) else length
 
     @classmethod
     def maybe_other_type(cls, dtype):
@@ -65,6 +48,7 @@ class ArgMin(Reducer):
         dtype = cls.maybe_other_type(array.dtype)
         result = array.nplike.empty(outlength, dtype=np.int64)
         if array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_argmin_complex",
@@ -74,12 +58,13 @@ class ArgMin(Reducer):
                 ](
                     result,
                     array.data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_argmin",
@@ -89,8 +74,8 @@ class ArgMin(Reducer):
                 ](
                     result,
                     array.data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
@@ -108,6 +93,7 @@ class ArgMax(Reducer):
         dtype = cls.maybe_other_type(array.dtype)
         result = array.nplike.empty(outlength, dtype=np.int64)
         if array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_argmax_complex",
@@ -116,13 +102,14 @@ class ArgMax(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_argmax",
@@ -131,9 +118,9 @@ class ArgMax(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
@@ -148,13 +135,14 @@ class Count(Reducer):
     def apply(cls, array, parents, outlength):
         assert isinstance(array, ak._v2.contents.NumpyArray)
         result = array.nplike.empty(outlength, dtype=np.int64)
+        assert parents.nplike is array.nplike
         array._handle_error(
             array.nplike[
                 "awkward_reduce_count_64", result.dtype.type, parents.dtype.type
             ](
                 result,
-                parents.to(array.nplike),
-                len(parents),
+                parents.data,
+                parents.length,
                 outlength,
             )
         )
@@ -171,6 +159,7 @@ class CountNonzero(Reducer):
         dtype = np.dtype(np.int64) if array.dtype.kind.upper() == "M" else array.dtype
         result = array.nplike.empty(outlength, dtype=np.int64)
         if array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_countnonzero_complex",
@@ -179,13 +168,14 @@ class CountNonzero(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_countnonzero",
@@ -194,9 +184,9 @@ class CountNonzero(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
@@ -211,9 +201,7 @@ class Sum(Reducer):
     def apply(cls, array, parents, outlength):
         assert isinstance(array, ak._v2.contents.NumpyArray)
         if array.dtype.kind == "M":
-            raise ValueError(
-                "cannot compute the sum (ak.sum) of {0}".format(repr(array.dtype))
-            )
+            raise ValueError(f"cannot compute the sum (ak.sum) of {array.dtype!r}")
         else:
             dtype = cls.maybe_other_type(array.dtype)
         result = array.nplike.empty(
@@ -223,6 +211,8 @@ class Sum(Reducer):
 
         if array.dtype == np.bool_:
             if result.dtype in (np.int64, np.uint64):
+                assert parents.nplike is array.nplike
+                assert parents.nplike is array.nplike
                 array._handle_error(
                     array.nplike[
                         "awkward_reduce_sum_int64_bool_64",
@@ -231,13 +221,14 @@ class Sum(Reducer):
                         parents.dtype.type,
                     ](
                         result,
-                        array._data,
-                        parents.to(array.nplike),
-                        len(parents),
+                        array.data,
+                        parents.data,
+                        parents.length,
                         outlength,
                     )
                 )
             elif result.dtype in (np.int32, np.uint32):
+                assert parents.nplike is array.nplike
                 array._handle_error(
                     array.nplike[
                         "awkward_reduce_sum_int32_bool_64",
@@ -246,15 +237,16 @@ class Sum(Reducer):
                         parents.dtype.type,
                     ](
                         result,
-                        array._data,
-                        parents.to(array.nplike),
-                        len(parents),
+                        array.data,
+                        parents.data,
+                        parents.length,
                         outlength,
                     )
                 )
             else:
                 raise NotImplementedError
         elif array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_sum_complex",
@@ -263,13 +255,14 @@ class Sum(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_sum",
@@ -278,9 +271,9 @@ class Sum(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
@@ -301,14 +294,13 @@ class Prod(Reducer):
     def apply(cls, array, parents, outlength):
         assert isinstance(array, ak._v2.contents.NumpyArray)
         if array.dtype.kind.upper() == "M":
-            raise ValueError(
-                "cannot compute the product (ak.prod) of {0}".format(repr(array.dtype))
-            )
+            raise ValueError(f"cannot compute the product (ak.prod) of {array.dtype!r}")
         result = array.nplike.empty(
             cls.maybe_double_length(array.dtype.type, outlength),
             dtype=cls.return_dtype(array.dtype),
         )
         if array.dtype == np.bool_:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_prod_bool",
@@ -317,13 +309,14 @@ class Prod(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         elif array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_prod_complex",
@@ -332,13 +325,14 @@ class Prod(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_prod",
@@ -347,9 +341,9 @@ class Prod(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
@@ -369,6 +363,7 @@ class Any(Reducer):
         dtype = cls.maybe_other_type(array.dtype)
         result = array.nplike.empty(outlength, dtype=np.bool_)
         if array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_sum_bool_complex",
@@ -377,13 +372,14 @@ class Any(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_sum_bool",
@@ -392,9 +388,9 @@ class Any(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
@@ -411,6 +407,7 @@ class All(Reducer):
         dtype = cls.maybe_other_type(array.dtype)
         result = array.nplike.empty(outlength, dtype=np.bool_)
         if array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_prod_bool_complex",
@@ -419,13 +416,14 @@ class All(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_prod_bool",
@@ -434,9 +432,9 @@ class All(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
@@ -457,15 +455,15 @@ class Min(Reducer):
     @staticmethod
     def _min_initial(initial, type):
         if initial is None:
-            if (
-                type == np.int8
-                or type == np.int16
-                or type == np.int32
-                or type == np.int64
-                or type == np.uint8
-                or type == np.uint16
-                or type == np.uint32
-                or type == np.uint64
+            if type in (
+                np.int8,
+                np.int16,
+                np.int32,
+                np.int64,
+                np.uint8,
+                np.uint16,
+                np.uint32,
+                np.uint64,
             ):
                 return np.iinfo(type).max
             else:
@@ -481,6 +479,7 @@ class Min(Reducer):
             cls.maybe_double_length(array.dtype.type, outlength), dtype=dtype
         )
         if array.dtype == np.bool_:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_prod_bool",
@@ -489,13 +488,14 @@ class Min(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         elif array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_min_complex",
@@ -504,14 +504,15 @@ class Min(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                     cls._min_initial(cls.initial, dtype),
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_min",
@@ -520,9 +521,9 @@ class Min(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                     cls._min_initial(cls.initial, dtype),
                 )
@@ -549,15 +550,15 @@ class Max(Reducer):
     @staticmethod
     def _max_initial(initial, type):
         if initial is None:
-            if (
-                type == np.int8
-                or type == np.int16
-                or type == np.int32
-                or type == np.int64
-                or type == np.uint8
-                or type == np.uint16
-                or type == np.uint32
-                or type == np.uint64
+            if type in (
+                np.int8,
+                np.int16,
+                np.int32,
+                np.int64,
+                np.uint8,
+                np.uint16,
+                np.uint32,
+                np.uint64,
             ):
                 return np.iinfo(type).min
             else:
@@ -573,6 +574,7 @@ class Max(Reducer):
             cls.maybe_double_length(array.dtype.type, outlength), dtype=dtype
         )
         if array.dtype == np.bool_:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_sum_bool",
@@ -581,13 +583,14 @@ class Max(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                 )
             )
         elif array.dtype.type in (np.complex128, np.complex64):
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_max_complex",
@@ -596,14 +599,15 @@ class Max(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                     cls._max_initial(cls.initial, dtype),
                 )
             )
         else:
+            assert parents.nplike is array.nplike
             array._handle_error(
                 array.nplike[
                     "awkward_reduce_max",
@@ -612,9 +616,9 @@ class Max(Reducer):
                     parents.dtype.type,
                 ](
                     result,
-                    array._data,
-                    parents.to(array.nplike),
-                    len(parents),
+                    array.data,
+                    parents.data,
+                    parents.length,
                     outlength,
                     cls._max_initial(cls.initial, dtype),
                 )
