@@ -1,7 +1,5 @@
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
-from __future__ import absolute_import
-
 import json
 import copy
 
@@ -19,36 +17,46 @@ numpy = ak.nplike.Numpy.instance()
 class ByteMaskedArray(Content):
     is_OptionType = True
 
-    def __init__(self, mask, content, valid_when, identifier=None, parameters=None):
+    def __init__(
+        self, mask, content, valid_when, identifier=None, parameters=None, nplike=None
+    ):
         if not (isinstance(mask, Index) and mask.dtype == np.dtype(np.int8)):
             raise TypeError(
-                "{0} 'mask' must be an Index with dtype=int8, not {1}".format(
+                "{} 'mask' must be an Index with dtype=int8, not {}".format(
                     type(self).__name__, repr(mask)
                 )
             )
         if not isinstance(content, Content):
             raise TypeError(
-                "{0} 'content' must be a Content subtype, not {1}".format(
+                "{} 'content' must be a Content subtype, not {}".format(
                     type(self).__name__, repr(content)
                 )
             )
         if not isinstance(valid_when, bool):
             raise TypeError(
-                "{0} 'valid_when' must be boolean, not {1}".format(
+                "{} 'valid_when' must be boolean, not {}".format(
                     type(self).__name__, repr(valid_when)
                 )
             )
-        if not len(mask) <= len(content):
+        if (
+            mask.nplike.known_shape
+            and content.nplike.known_shape
+            and mask.length > content.length
+        ):
             raise ValueError(
-                "{0} len(mask) ({1}) must be <= len(content) ({2})".format(
-                    type(self).__name__, len(mask), len(content)
+                "{} len(mask) ({}) must be <= len(content) ({})".format(
+                    type(self).__name__, mask.length, content.length
                 )
             )
+        if nplike is None:
+            nplike = content.nplike
+        if nplike is None:
+            nplike = mask.nplike
 
         self._mask = mask
         self._content = content
         self._valid_when = valid_when
-        self._init(identifier, parameters)
+        self._init(identifier, parameters, nplike)
 
     @property
     def mask(self):
@@ -62,10 +70,6 @@ class ByteMaskedArray(Content):
     def valid_when(self):
         return self._valid_when
 
-    @property
-    def nplike(self):
-        return self._mask.nplike
-
     Form = ByteMaskedForm
 
     def _form_with_key(self, getkey):
@@ -82,21 +86,34 @@ class ByteMaskedArray(Content):
     def _to_buffers(self, form, getkey, container, nplike):
         assert isinstance(form, self.Form)
         key = getkey(self, form, "mask")
-        container[key] = ak._v2._util.little_endian(self._mask.to(nplike))
+        container[key] = ak._v2._util.little_endian(self._mask.raw(nplike))
         self._content._to_buffers(form.content, getkey, container, nplike)
 
     @property
     def typetracer(self):
+        tt = ak._v2._typetracer.TypeTracer.instance()
         return ByteMaskedArray(
-            ak._v2.index.Index(self._mask.to(ak._v2._typetracer.TypeTracer.instance())),
+            ak._v2.index.Index(self._mask.raw(tt)),
             self._content.typetracer,
             self._valid_when,
             self._typetracer_identifier(),
             self._parameters,
+            tt,
         )
 
-    def __len__(self):
-        return len(self._mask)
+    @property
+    def length(self):
+        return self._mask.length
+
+    def _forget_length(self):
+        return ByteMaskedArray(
+            self._mask.forget_length(),
+            self._content,
+            self._valid_when,
+            self._identifier,
+            self._parameters,
+            self._nplike,
+        )
 
     def __repr__(self):
         return self._repr("", "", "")
@@ -105,7 +122,7 @@ class ByteMaskedArray(Content):
         out = [indent, pre, "<ByteMaskedArray valid_when="]
         out.append(repr(json.dumps(self._valid_when)))
         out.append(" len=")
-        out.append(repr(str(len(self))))
+        out.append(repr(str(self.length)))
         out.append(">")
         out.extend(self._repr_extra(indent + "    "))
         out.append("\n")
@@ -122,46 +139,50 @@ class ByteMaskedArray(Content):
             self._valid_when,
             self._identifier,
             ak._v2._util.merge_parameters(self._parameters, parameters),
+            self._nplike,
         )
 
     def toIndexedOptionArray64(self):
-        nplike = self.nplike
-        index = ak._v2.index.Index64.empty(len(self._mask), nplike)
+        index = ak._v2.index.Index64.empty(self._mask.length, self._nplike)
+        assert index.nplike is self._nplike and self._mask.nplike is self._nplike
         self._handle_error(
-            nplike[
+            self._nplike[
                 "awkward_ByteMaskedArray_toIndexedOptionArray",
                 index.dtype.type,
                 self._mask.dtype.type,
             ](
-                index.to(nplike),
-                self._mask.to(nplike),
-                len(self._mask),
+                index.data,
+                self._mask.data,
+                self._mask.length,
                 self._valid_when,
             ),
         )
 
         return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
-            index, self._content, self._identifier, self._parameters
+            index, self._content, self._identifier, self._parameters, self._nplike
         )
 
     def mask_as_bool(self, valid_when=None, nplike=None):
         if valid_when is None:
             valid_when = self._valid_when
         if nplike is None:
-            nplike = self._mask.nplike
+            nplike = self._nplike
 
         if valid_when == self._valid_when:
-            return self._mask.to(nplike) != 0
+            return self._mask.raw(nplike) != 0
         else:
-            return self._mask.to(nplike) != 1
+            return self._mask.raw(nplike) != 1
 
     def _getitem_nothing(self):
         return self._content._getitem_range(slice(0, 0))
 
     def _getitem_at(self, where):
+        if not self._nplike.known_data:
+            return ak._v2._typetracer.MaybeNone(self._content._getitem_at(where))
+
         if where < 0:
-            where += len(self)
-        if not (0 <= where < len(self)) and self.nplike.known_shape:
+            where += self.length
+        if self._nplike.known_shape and not 0 <= where < self.length:
             raise NestedIndexError(self, where)
         if self._mask[where] == self._valid_when:
             return self._content._getitem_at(where)
@@ -169,7 +190,10 @@ class ByteMaskedArray(Content):
             return None
 
     def _getitem_range(self, where):
-        start, stop, step = where.indices(len(self))
+        if not self._nplike.known_shape:
+            return self
+
+        start, stop, step = where.indices(self.length)
         assert step == 1
         return ByteMaskedArray(
             self._mask[start:stop],
@@ -177,6 +201,7 @@ class ByteMaskedArray(Content):
             self._valid_when,
             self._range_identifier(start, stop),
             self._parameters,
+            self._nplike,
         )
 
     def _getitem_field(self, where, only_fields=()):
@@ -186,7 +211,8 @@ class ByteMaskedArray(Content):
             self._valid_when,
             self._field_identifier(where),
             None,
-        )
+            self._nplike,
+        ).simplify_optiontype()
 
     def _getitem_fields(self, where, only_fields=()):
         return ByteMaskedArray(
@@ -195,7 +221,8 @@ class ByteMaskedArray(Content):
             self._valid_when,
             self._fields_identifier(where),
             None,
-        )
+            self._nplike,
+        ).simplify_optiontype()
 
     def _carry(self, carry, allow_lazy, exception):
         assert isinstance(carry, ak._v2.index.Index)
@@ -214,59 +241,81 @@ class ByteMaskedArray(Content):
             self._valid_when,
             self._carry_identifier(carry, exception),
             self._parameters,
+            self._nplike,
         )
 
     def _nextcarry_outindex(self, numnull):
-        nplike = self.nplike
+        assert numnull.nplike is self._nplike and self._mask.nplike is self._nplike
         self._handle_error(
-            nplike[
+            self._nplike[
                 "awkward_ByteMaskedArray_numnull",
                 numnull.dtype.type,
                 self._mask.dtype.type,
             ](
-                numnull.to(nplike),
-                self._mask.to(nplike),
-                len(self._mask),
+                numnull.data,
+                self._mask.data,
+                self._mask.length,
                 self._valid_when,
             )
         )
-        nextcarry = ak._v2.index.Index64.empty(len(self) - numnull[0], nplike)
-        outindex = ak._v2.index.Index64.empty(len(self), nplike)
+        nextcarry = ak._v2.index.Index64.empty(self.length - numnull[0], self._nplike)
+        outindex = ak._v2.index.Index64.empty(self.length, self._nplike)
+        assert (
+            nextcarry.nplike is self._nplike
+            and outindex.nplike is self._nplike
+            and self._mask.nplike is self._nplike
+        )
         self._handle_error(
-            nplike[
+            self._nplike[
                 "awkward_ByteMaskedArray_getitem_nextcarry_outindex",
                 nextcarry.dtype.type,
                 outindex.dtype.type,
                 self._mask.dtype.type,
             ](
-                nextcarry.to(nplike),
-                outindex.to(nplike),
-                self._mask.to(nplike),
-                len(self._mask),
+                nextcarry.data,
+                outindex.data,
+                self._mask.raw(self._nplike),
+                self._mask.length,
                 self._valid_when,
             )
         )
         return nextcarry, outindex
 
     def _getitem_next_jagged_generic(self, slicestarts, slicestops, slicecontent, tail):
-        nplike = self.nplike
-        if len(slicestarts) != len(self) and nplike.known_shape:
+        if (
+            slicestarts.nplike.known_shape
+            and self._nplike.known_shape
+            and slicestarts.length != self.length
+        ):
             raise NestedIndexError(
                 self,
-                ak._v2.contents.ListArray(slicestarts, slicestops, slicecontent),
-                "cannot fit jagged slice with length {0} into {1} of size {2}".format(
-                    len(slicestarts), type(self).__name__, len(self)
+                ak._v2.contents.ListArray(
+                    slicestarts, slicestops, slicecontent, None, None, self._nplike
+                ),
+                "cannot fit jagged slice with length {} into {} of size {}".format(
+                    slicestarts.length, type(self).__name__, self.length
                 ),
             )
 
-        numnull = ak._v2.index.Index64.empty(1, nplike)
+        numnull = ak._v2.index.Index64.empty(1, self._nplike)
         nextcarry, outindex = self._nextcarry_outindex(numnull)
 
-        reducedstarts = ak._v2.index.Index64.empty(len(self) - numnull[0], nplike)
-        reducedstops = ak._v2.index.Index64.empty(len(self) - numnull[0], nplike)
+        reducedstarts = ak._v2.index.Index64.empty(
+            self.length - numnull[0], self._nplike
+        )
+        reducedstops = ak._v2.index.Index64.empty(
+            self.length - numnull[0], self._nplike
+        )
 
+        assert (
+            outindex.nplike is self._nplike
+            and slicestarts.nplike is self._nplike
+            and slicestops.nplike is self._nplike
+            and reducedstarts.nplike is self._nplike
+            and reducedstops.nplike is self._nplike
+        )
         self._handle_error(
-            nplike[
+            self._nplike[
                 "awkward_MaskedArray_getitem_next_jagged_project",
                 outindex.dtype.type,
                 slicestarts.dtype.type,
@@ -274,12 +323,12 @@ class ByteMaskedArray(Content):
                 reducedstarts.dtype.type,
                 reducedstops.dtype.type,
             ](
-                outindex.to(nplike),
-                slicestarts.to(nplike),
-                slicestops.to(nplike),
-                reducedstarts.to(nplike),
-                reducedstops.to(nplike),
-                len(self),
+                outindex.data,
+                slicestarts.data,
+                slicestops.data,
+                reducedstarts.data,
+                reducedstops.data,
+                self.length,
             )
         )
 
@@ -287,7 +336,7 @@ class ByteMaskedArray(Content):
         out = next._getitem_next_jagged(reducedstarts, reducedstops, slicecontent, tail)
 
         out2 = ak._v2.contents.indexedoptionarray.IndexedOptionArray(
-            outindex, out, self._identifier, self._parameters
+            outindex, out, self._identifier, self._parameters, self._nplike
         )
         return out2.simplify_optiontype()
 
@@ -302,7 +351,7 @@ class ByteMaskedArray(Content):
 
         elif isinstance(head, (int, slice, ak._v2.index.Index64)):
             nexthead, nexttail = ak._v2._slicing.headtail(tail)
-            numnull = ak._v2.index.Index64.empty(1, self.nplike)
+            numnull = ak._v2.index.Index64.empty(1, self._nplike)
             nextcarry, outindex = self._nextcarry_outindex(numnull)
 
             next = self._content._carry(nextcarry, True, NestedIndexError)
@@ -312,6 +361,7 @@ class ByteMaskedArray(Content):
                 out,
                 self._identifier,
                 self._parameters,
+                self._nplike,
             )
             return out2.simplify_optiontype()
 
@@ -337,62 +387,76 @@ class ByteMaskedArray(Content):
             raise AssertionError(repr(head))
 
     def project(self, mask=None):
-        mask_length = len(self._mask)
-        numnull = ak._v2.index.Index64.zeros(1, self.nplike)
+        mask_length = self._mask.length
+        numnull = ak._v2.index.Index64.zeros(1, self._nplike)
 
         if mask is not None:
-            if mask_length != len(mask):
+            if self._nplike.known_shape and mask_length != mask.length:
                 raise ValueError(
-                    "mask length ({0}) is not equal to {1} length ({2})".format(
-                        len(mask), type(self).__name__, mask_length
+                    "mask length ({}) is not equal to {} length ({})".format(
+                        mask.length, type(self).__name__, mask_length
                     )
                 )
 
-            nextmask = ak._v2.index.Index8.zeros(mask_length, self.nplike)
+            nextmask = ak._v2.index.Index8.empty(mask_length, self._nplike)
+            assert (
+                nextmask.nplike is self._nplike
+                and mask.nplike is self._nplike
+                and self._mask.nplike is self._nplike
+            )
             self._handle_error(
-                self.nplike[
+                self._nplike[
                     "awkward_ByteMaskedArray_overlay_mask",
                     nextmask.dtype.type,
                     mask.dtype.type,
                     self._mask.dtype.type,
                 ](
-                    nextmask.to(self.nplike),
-                    mask.to(self.nplike),
-                    self._mask.to(self.nplike),
+                    nextmask.data,
+                    mask.data,
+                    self._mask.data,
                     mask_length,
                     self._valid_when,
                 )
             )
             valid_when = False
             next = ByteMaskedArray(
-                nextmask, self._content, valid_when, self._identifier, self._parameters
+                nextmask,
+                self._content,
+                valid_when,
+                self._identifier,
+                self._parameters,
+                self._nplike,
             )
             return next.project()
 
         else:
+            assert numnull.nplike is self._nplike and self._mask.nplike is self._nplike
             self._handle_error(
                 self.nplike[
                     "awkward_ByteMaskedArray_numnull",
                     numnull.dtype.type,
                     self._mask.dtype.type,
                 ](
-                    numnull.to(self.nplike),
-                    self._mask.to(self.nplike),
+                    numnull.data,
+                    self._mask.data,
                     mask_length,
                     self._valid_when,
                 )
             )
-            nextcarry = ak._v2.index.Index64.zeros(
-                mask_length - numnull[0], self.nplike
+            nextcarry = ak._v2.index.Index64.empty(
+                mask_length - numnull[0], self._nplike
+            )
+            assert (
+                nextcarry.nplike is self._nplike and self._mask.nplike is self._nplike
             )
             self._handle_error(
-                self.nplike[
+                self._nplike[
                     "awkward_ByteMaskedArray_getitem_nextcarry",
                     nextcarry.dtype.type,
                     self._mask.dtype.type,
                 ](
-                    nextcarry.to(self.nplike),
-                    self._mask.to(self.nplike),
+                    nextcarry.data,
+                    self._mask.data,
                     mask_length,
                     self._valid_when,
                 )
@@ -407,29 +471,31 @@ class ByteMaskedArray(Content):
                 ak._v2.contents.indexedarray.IndexedArray,
                 ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                 ak._v2.contents.bytemaskedarray.ByteMaskedArray,
-                ak._v2.contents.bitmaskeddarray.BitMaskedArray,
-                ak._v2.contents.unmaskeddarray.UnmaskedArray,
+                ak._v2.contents.bitmaskedarray.BitMaskedArray,
+                ak._v2.contents.unmaskedarray.UnmaskedArray,
             ),
         ):
-            return self.toIndexedOptionArray64.simplify_optiontype
+            return self.toIndexedOptionArray64().simplify_optiontype()
         else:
             return self
 
     def num(self, axis, depth=0):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            out = ak._v2.index.Index64.empty(1, self.nplike)
-            out[0] = len(self)
-            return ak._v2.contents.numpyarray.NumpyArray(out)[0]
+            out = ak._v2.index.Index64.empty(1, self._nplike)
+            out[0] = self.length
+            return ak._v2.contents.numpyarray.NumpyArray(out, None, None, self._nplike)[
+                0
+            ]
         else:
-            numnull = ak._v2.index.Index64.empty(1, self.nplike, dtype=np.int64)
+            numnull = ak._v2.index.Index64.empty(1, self._nplike, dtype=np.int64)
             nextcarry, outindex = self._nextcarry_outindex(numnull)
 
             next = self._content._carry(nextcarry, False, NestedIndexError)
             out = next.num(posaxis, depth)
 
             out2 = ak._v2.contents.indexedoptionarray.IndexedOptionArray(
-                outindex, out, parameters=self.parameters
+                outindex, out, None, self.parameters, self._nplike
             )
             return out2.simplify_optiontype()
 
@@ -438,38 +504,43 @@ class ByteMaskedArray(Content):
         if posaxis == depth:
             raise np.AxisError(self, "axis=0 not allowed for flatten")
         else:
-            numnull = ak._v2.index.Index64.empty(1, self.nplike)
+            numnull = ak._v2.index.Index64.empty(1, self._nplike)
             nextcarry, outindex = self._nextcarry_outindex(numnull)
 
             next = self._content._carry(nextcarry, False, NestedIndexError)
 
             offsets, flattened = next._offsets_and_flattened(posaxis, depth)
 
-            if len(offsets) == 0:
+            if offsets.length == 0:
                 return (
                     offsets,
                     ak._v2.contents.indexedoptionarray.IndexedOptionArray(
-                        outindex, flattened, None, self._parameters
+                        outindex, flattened, None, self._parameters, self._nplike
                     ),
                 )
 
             else:
                 outoffsets = ak._v2.index.Index64.empty(
-                    len(offsets) + numnull[0], self.nplike, dtype=np.int64
+                    offsets.length + numnull[0], self._nplike, dtype=np.int64
                 )
 
+                assert (
+                    outoffsets.nplike is self._nplike
+                    and outindex.nplike is self._nplike
+                    and offsets.nplike is self._nplike
+                )
                 self._handle_error(
-                    self.nplike[
+                    self._nplike[
                         "awkward_IndexedArray_flatten_none2empty",
                         outoffsets.dtype.type,
                         outindex.dtype.type,
                         offsets.dtype.type,
                     ](
-                        outoffsets.to(self.nplike),
-                        outindex.to(self.nplike),
-                        len(outindex),
-                        offsets.to(self.nplike),
-                        len(offsets),
+                        outoffsets.data,
+                        outindex.data,
+                        outindex.length,
+                        offsets.data,
+                        offsets.length,
                     )
                 )
                 return (outoffsets, flattened)
@@ -497,7 +568,7 @@ class ByteMaskedArray(Content):
                 ak._v2.contents.unmaskedarray.UnmaskedArray,
             ),
         ):
-            self._content.mergeable(other.content, mergebool)
+            return self._content.mergeable(other.content, mergebool)
 
         else:
             return self._content.mergeable(other, mergebool)
@@ -518,7 +589,7 @@ class ByteMaskedArray(Content):
         if posaxis == depth:
             return self._localindex_axis0()
         else:
-            numnull = ak._v2.index.Index64.empty(1, self.nplike)
+            numnull = ak._v2.index.Index64.empty(1, self._nplike)
             nextcarry, outindex = self._nextcarry_outindex(numnull)
 
             next = self._content._carry(nextcarry, False, NestedIndexError)
@@ -528,6 +599,7 @@ class ByteMaskedArray(Content):
                 out,
                 self._identifier,
                 self._parameters,
+                self._nplike,
             )
             return out2.simplify_optiontype()
 
@@ -538,17 +610,18 @@ class ByteMaskedArray(Content):
             self._valid_when,
             self._identifier,
             self._parameters,
+            self._nplike,
         )
 
     def _is_unique(self, negaxis, starts, parents, outlength):
-        if len(self._mask) == 0:
+        if self._mask.length == 0:
             return True
         return self.toIndexedOptionArray64()._is_unique(
             negaxis, starts, parents, outlength
         )
 
     def _unique(self, negaxis, starts, parents, outlength):
-        if len(self._mask) == 0:
+        if self._mask.length == 0:
             return self
         return self.toIndexedOptionArray64()._unique(
             negaxis, starts, parents, outlength
@@ -566,8 +639,10 @@ class ByteMaskedArray(Content):
         kind,
         order,
     ):
-        if len(self._mask) == 0:
-            return ak._v2.contents.NumpyArray(self.nplike.empty(0, np.int64))
+        if self._mask.length == 0:
+            return ak._v2.contents.NumpyArray(
+                self._nplike.empty(0, np.int64), None, None, self._nplike
+            )
 
         return self.toIndexedOptionArray64()._argsort_next(
             negaxis,
@@ -602,7 +677,7 @@ class ByteMaskedArray(Content):
         if posaxis == depth:
             return self._combinations_axis0(n, replacement, recordlookup, parameters)
         else:
-            numnull = ak._v2.index.Index64.empty(1, self.nplike, dtype=np.int64)
+            numnull = ak._v2.index.Index64.empty(1, self._nplike, dtype=np.int64)
             nextcarry, outindex = self._nextcarry_outindex(numnull)
 
             next = self._content._carry(nextcarry, True, NestedIndexError)
@@ -610,7 +685,7 @@ class ByteMaskedArray(Content):
                 n, replacement, recordlookup, parameters, posaxis, depth
             )
             out2 = ak._v2.contents.indexedoptionarray.IndexedOptionArray(
-                outindex, out, parameters=parameters
+                outindex, out, None, parameters, self._nplike
             )
             return out2.simplify_optiontype()
 
@@ -625,30 +700,36 @@ class ByteMaskedArray(Content):
         mask,
         keepdims,
     ):
-        nplike = self.nplike
+        mask_length = self._mask.length
 
-        mask_length = len(self._mask)
-
-        numnull = ak._v2.index.Index64.zeros(1, nplike)
+        numnull = ak._v2.index.Index64.empty(1, self._nplike)
+        assert numnull.nplike is self._nplike and self._mask.nplike is self._nplike
         self._handle_error(
-            nplike[
+            self._nplike[
                 "awkward_ByteMaskedArray_numnull",
                 numnull.dtype.type,
                 self._mask.dtype.type,
             ](
-                numnull.to(nplike),
-                self._mask.to(nplike),
+                numnull.data,
+                self._mask.data,
                 mask_length,
                 self._valid_when,
             )
         )
 
         next_length = mask_length - numnull[0]
-        nextcarry = ak._v2.index.Index64.zeros(next_length, nplike)
-        nextparents = ak._v2.index.Index64.zeros(next_length, nplike)
-        outindex = ak._v2.index.Index64.zeros(mask_length, nplike)
+        nextcarry = ak._v2.index.Index64.empty(next_length, self._nplike)
+        nextparents = ak._v2.index.Index64.empty(next_length, self._nplike)
+        outindex = ak._v2.index.Index64.empty(mask_length, self._nplike)
+        assert (
+            nextcarry.nplike is self._nplike
+            and nextparents.nplike is self._nplike
+            and outindex.nplike is self._nplike
+            and self._mask.nplike is self._nplike
+            and parents.nplike is self._nplike
+        )
         self._handle_error(
-            nplike[
+            self._nplike[
                 "awkward_ByteMaskedArray_reduce_next_64",
                 nextcarry.dtype.type,
                 nextparents.dtype.type,
@@ -656,11 +737,11 @@ class ByteMaskedArray(Content):
                 self._mask.dtype.type,
                 parents.dtype.type,
             ](
-                nextcarry.to(nplike),
-                nextparents.to(nplike),
-                outindex.to(nplike),
-                self._mask.to(nplike),
-                parents.to(nplike),
+                nextcarry.data,
+                nextparents.data,
+                outindex.data,
+                self._mask.data,
+                parents.data,
                 mask_length,
                 self._valid_when,
             )
@@ -669,33 +750,41 @@ class ByteMaskedArray(Content):
         branch, depth = self.branch_depth
 
         if reducer.needs_position and (not branch and negaxis == depth):
-            nextshifts = ak._v2.index.Index64.zeros(next_length, nplike)
+            nextshifts = ak._v2.index.Index64.empty(next_length, self._nplike)
             if shifts is None:
+                assert (
+                    nextshifts.nplike is self._nplike
+                    and self._mask.nplike is self._nplike
+                )
                 self._handle_error(
-                    nplike[
+                    self._nplike[
                         "awkward_ByteMaskedArray_reduce_next_nonlocal_nextshifts_64",
                         nextshifts.dtype.type,
                         self._mask.dtype.type,
                     ](
-                        nextshifts.to(nplike),
-                        self._mask.to(nplike),
+                        nextshifts.data,
+                        self._mask.data,
                         mask_length,
                         self._valid_when,
                     )
                 )
             else:
+                assert (
+                    nextshifts.nplike is self._nplike
+                    and self._mask.nplike is self._nplike
+                )
                 self._handle_error(
-                    nplike[
+                    self._nplike[
                         "awkward_ByteMaskedArray_reduce_next_nonlocal_nextshifts_fromshifts_64",
                         nextshifts.dtype.type,
                         self._mask.dtype.type,
                         shifts.dtype.type,
                     ](
-                        nextshifts.to(nplike),
-                        self._mask.to(nplike),
+                        nextshifts.data,
+                        self._mask.data,
                         mask_length,
                         self._valid_when,
-                        shifts.to(nplike),
+                        shifts.data,
                     )
                 )
         else:
@@ -723,17 +812,18 @@ class ByteMaskedArray(Content):
                 out = out.toListOffsetArray64(True)
 
             if isinstance(out, ak._v2.contents.ListOffsetArray):
-                outoffsets = ak._v2.index.Index64.zeros(len(starts) + 1, nplike)
+                outoffsets = ak._v2.index.Index64.empty(starts.length + 1, self._nplike)
+                assert outoffsets.nplike is self._nplike
                 self._handle_error(
-                    nplike[
+                    self._nplike[
                         "awkward_IndexedArray_reduce_next_fix_offsets_64",
                         outoffsets.dtype.type,
                         starts.dtype.type,
                     ](
-                        outoffsets.to(nplike),
-                        starts.to(nplike),
-                        len(starts),
-                        len(outindex),
+                        outoffsets.data,
+                        starts.data,
+                        starts.length,
+                        outindex.length,
                     )
                 )
 
@@ -742,6 +832,7 @@ class ByteMaskedArray(Content):
                     out.content,
                     None,
                     None,
+                    self._nplike,
                 ).simplify_optiontype()
 
                 return ak._v2.contents.ListOffsetArray(
@@ -749,6 +840,7 @@ class ByteMaskedArray(Content):
                     tmp,
                     None,
                     None,
+                    self._nplike,
                 )
 
             else:
@@ -759,8 +851,8 @@ class ByteMaskedArray(Content):
                 )
 
     def _validityerror(self, path):
-        if len(self._content) < len(self.mask):
-            return 'at {0} ("{1}"): len(content) < len(mask)'.format(path, type(self))
+        if self._nplike.known_shape and self._content.length < self.mask.length:
+            return f'at {path} ("{type(self)}"): len(content) < len(mask)'
         elif isinstance(
             self._content,
             (
@@ -785,16 +877,17 @@ class ByteMaskedArray(Content):
         if not self._valid_when:
             return self._mask
         else:
-            out = ak._v2.index.Index64.empty(len(self), self.nplike)
+            out = ak._v2.index.Index64.empty(self.length, self._nplike)
+            assert out.nplike is self._nplike and self._mask.nplike is self._nplike
             self._handle_error(
-                self.nplike[
+                self._nplike[
                     "awkward_ByteMaskedArray_mask",
                     out.dtype.type,
                     self._mask.dtype.type,
                 ](
-                    out.to(self.nplike),
-                    self._mask.to(self.nplike),
-                    len(self._mask),
+                    out.data,
+                    self._mask.data,
+                    self._mask.length,
                     self._valid_when,
                 )
             )
@@ -806,25 +899,27 @@ class ByteMaskedArray(Content):
             return self.rpad_axis0(target, clip)
         elif posaxis == depth + 1:
             mask = self.bytemask()
-            index = ak._v2.index.Index64.empty(len(mask), self.nplike)
+            index = ak._v2.index.Index64.empty(mask.length, self._nplike)
+            assert index.nplike is self._nplike and self._mask.nplike is self._nplike
             self._handle_error(
-                self.nplike[
+                self._nplike[
                     "awkward_IndexedOptionArray_rpad_and_clip_mask_axis1",
                     index.dtype.type,
                     self._mask.dtype.type,
                 ](
-                    index.to(self.nplike),
-                    self._mask.to(self.nplike),
-                    len(self._mask),
+                    index.data,
+                    self._mask.data,
+                    self._mask.length,
                 )
             )
             next = self.project()._rpad(target, posaxis, depth, clip)
             return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
                 index,
-                next.simplify_optiontype(),
+                next,
                 None,
                 self._parameters,
-            )
+                self._nplike,
+            ).simplify_optiontype()
         else:
             return ak._v2.contents.bytemaskedarray.ByteMaskedArray(
                 self._mask,
@@ -832,6 +927,7 @@ class ByteMaskedArray(Content):
                 self._valid_when,
                 None,
                 self._parameters,
+                self._nplike,
             )
 
     def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
@@ -869,6 +965,7 @@ class ByteMaskedArray(Content):
                     self._valid_when,
                     self._identifier,
                     self._parameters if options["keep_parameters"] else None,
+                    self._nplike,
                 )
 
         else:
@@ -902,20 +999,21 @@ class ByteMaskedArray(Content):
         if self._content.is_RecordType:
             next = self.toIndexedOptionArray64()
             content = next._content.packed()
-            if len(content) > len(self._mask):
-                content = content[: len(self._mask)]
+            if content.length > self._mask.length:
+                content = content[: self._mask.length]
 
             return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
                 next._index,
                 content,
                 next._identifier,
                 next._parameters,
+                self._nplike,
             )
 
         else:
             content = self._content.packed()
-            if len(content) > len(self._mask):
-                content = content[: len(self._mask)]
+            if content.length > self._mask.length:
+                content = content[: self._mask.length]
 
             return ByteMaskedArray(
                 self._mask,
@@ -923,6 +1021,7 @@ class ByteMaskedArray(Content):
                 self._valid_when,
                 self._identifier,
                 self._parameters,
+                self._nplike,
             )
 
     def _to_list(self, behavior):
@@ -930,9 +1029,49 @@ class ByteMaskedArray(Content):
         if out is not None:
             return out
 
-        mask = self.mask_as_bool(valid_when=True, nplike=numpy)
+        mask = self.mask_as_bool(valid_when=True, nplike=self.nplike)
         content = self._content._to_list(behavior)
-        out = [None] * len(self._mask)
+        out = [None] * self._mask.length
+        for i, isvalid in enumerate(mask):
+            if isvalid:
+                out[i] = content[i]
+        return out
+
+    def _to_nplike(self, nplike):
+        content = self._content._to_nplike(nplike)
+        mask = self._mask._to_nplike(nplike)
+        return ByteMaskedArray(
+            mask,
+            content,
+            valid_when=self._valid_when,
+            identifier=self._identifier,
+            parameters=self._parameters,
+            nplike=nplike,
+        )
+
+    def _to_json(
+        self,
+        behavior,
+        nan_string,
+        infinity_string,
+        minus_infinity_string,
+        complex_real_string,
+        complex_imag_string,
+    ):
+        out = self._to_list_custom(behavior)
+        if out is not None:
+            return out
+
+        mask = self.mask_as_bool(valid_when=True, nplike=self.nplike)
+        content = self._content._to_json(
+            behavior,
+            nan_string,
+            infinity_string,
+            minus_infinity_string,
+            complex_real_string,
+            complex_imag_string,
+        )
+        out = [None] * self._mask.length
         for i, isvalid in enumerate(mask):
             if isvalid:
                 out[i] = content[i]
