@@ -1,14 +1,9 @@
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
-from __future__ import absolute_import
-
 import json
 import copy
 
-try:
-    from collections.abc import Iterable
-except ImportError:
-    from collections import Iterable
+from collections.abc import Iterable
 
 import awkward as ak
 from awkward._v2.record import Record
@@ -18,15 +13,24 @@ from awkward._v2.forms.recordform import RecordForm
 from awkward._v2.forms.form import _parameters_equal
 
 np = ak.nplike.NumpyMetadata.instance()
+numpy = ak.nplike.Numpy.instance()
 
 
 class RecordArray(Content):
     is_RecordType = True
 
-    def __init__(self, contents, fields, length=None, identifier=None, parameters=None):
+    def __init__(
+        self,
+        contents,
+        fields,
+        length=None,
+        identifier=None,
+        parameters=None,
+        nplike=None,
+    ):
         if not isinstance(contents, Iterable):
             raise TypeError(
-                "{0} 'contents' must be iterable, not {1}".format(
+                "{} 'contents' must be iterable, not {}".format(
                     type(self).__name__, repr(contents)
                 )
             )
@@ -35,29 +39,37 @@ class RecordArray(Content):
 
         if len(contents) == 0 and length is None:
             raise TypeError(
-                "{0} if len(contents) == 0, a 'length' must be specified".format(
+                "{} if len(contents) == 0, a 'length' must be specified".format(
                     type(self).__name__
                 )
             )
         elif length is None:
-            length = min(len(x) for x in contents)
-        if not (ak._util.isint(length) and length >= 0):
+            lengths = [x.length for x in contents]
+            if any(
+                isinstance(x, ak._v2._typetracer.UnknownLengthType) for x in contents
+            ):
+                length = ak._v2._typetracer.UnknownLength
+            else:
+                length = min(lengths)
+        if not isinstance(length, ak._v2._typetracer.UnknownLengthType) and not (
+            ak._util.isint(length) and length >= 0
+        ):
             raise TypeError(
-                "{0} 'length' must be a non-negative integer or None, not {1}".format(
+                "{} 'length' must be a non-negative integer or None, not {}".format(
                     type(self).__name__, repr(length)
                 )
             )
         for content in contents:
             if not isinstance(content, Content):
                 raise TypeError(
-                    "{0} all 'contents' must be Content subclasses, not {1}".format(
+                    "{} all 'contents' must be Content subclasses, not {}".format(
                         type(self).__name__, repr(content)
                     )
                 )
-            if not len(content) >= length:
+            if content.length < length:
                 raise ValueError(
-                    "{0} len(content) ({1}) must be <= length ({2}) for all 'contents'".format(
-                        type(self).__name__, len(content), length
+                    "{} len(content) ({}) must be >= length ({}) for all 'contents'".format(
+                        type(self).__name__, content.length, length
                     )
                 )
 
@@ -66,27 +78,42 @@ class RecordArray(Content):
                 fields = list(fields)
             if not all(ak._util.isstr(x) for x in fields):
                 raise TypeError(
-                    "{0} 'fields' must all be strings, not {1}".format(
+                    "{} 'fields' must all be strings, not {}".format(
                         type(self).__name__, repr(fields)
                     )
                 )
             if not len(contents) == len(fields):
                 raise ValueError(
-                    "{0} len(contents) ({1}) must be equal to len(fields) ({2})".format(
+                    "{} len(contents) ({}) must be equal to len(fields) ({})".format(
                         type(self).__name__, len(contents), len(fields)
                     )
                 )
         elif fields is not None:
             raise TypeError(
-                "{0} 'fields' must be iterable or None, not {1}".format(
+                "{} 'fields' must be iterable or None, not {}".format(
                     type(self).__name__, repr(fields)
                 )
             )
+        if nplike is None:
+            for content in contents:
+                if nplike is None:
+                    nplike = content.nplike
+                    break
+                elif nplike is not content.nplike:
+                    raise TypeError(
+                        "{} 'contents' must use the same array library (nplike): {} vs {}".format(
+                            type(self).__name__,
+                            type(nplike).__name__,
+                            type(content.nplike).__name__,
+                        )
+                    )
+        if nplike is None:
+            nplike = numpy
 
         self._contents = contents
         self._fields = fields
         self._length = length
-        self._init(identifier, parameters)
+        self._init(identifier, parameters, nplike)
 
     @property
     def contents(self):
@@ -108,18 +135,12 @@ class RecordArray(Content):
         return RecordArray(
             self._contents,
             None,
-            length=self._length,
-            identifier=None,
-            parameters=None,
+            self._length,
+            None,
+            None,
+            self._nplike,
         )
 
-    @property
-    def nplike(self):
-        for content in self._contents:
-            return content.nplike
-        else:
-            return ak.nplike.Numpy.instance()
-
     Form = RecordForm
 
     def _form_with_key(self, getkey):
@@ -143,17 +164,30 @@ class RecordArray(Content):
 
     @property
     def typetracer(self):
+        tt = ak._v2._typetracer.TypeTracer.instance()
         return RecordArray(
             [x.typetracer for x in self._contents],
             self._fields,
             self._length,
             self._typetracer_identifier(),
             self._parameters,
+            tt,
         )
 
-    def __len__(self):
+    @property
+    def length(self):
         return self._length
 
+    def _forget_length(self):
+        return RecordArray(
+            [x._forget_length() for x in self._contents],
+            self._fields,
+            ak._v2._typetracer.UnknownLength,
+            self._identifier,
+            self._parameters,
+            self._nplike,
+        )
+
     def __repr__(self):
         return self._repr("", "", "")
 
@@ -161,20 +195,20 @@ class RecordArray(Content):
         out = [indent, pre, "<RecordArray is_tuple="]
         out.append(repr(json.dumps(self.is_tuple)))
         out.append(" len=")
-        out.append(repr(str(len(self))))
+        out.append(repr(str(self.length)))
         out.append(">")
         out.extend(self._repr_extra(indent + "    "))
         out.append("\n")
 
         if self._fields is None:
             for i, x in enumerate(self._contents):
-                out.append("{0}    <content index={1}>\n".format(indent, repr(str(i))))
+                out.append(f"{indent}    <content index={repr(str(i))}>\n")
                 out.append(x._repr(indent + "        ", "", "\n"))
                 out.append(indent + "    </content>\n")
         else:
             for i, x in enumerate(self._contents):
                 out.append(
-                    "{0}    <content index={1} field={2}>\n".format(
+                    "{}    <content index={} field={}>\n".format(
                         indent, repr(str(i)), repr(self._fields[i])
                     )
                 )
@@ -192,6 +226,7 @@ class RecordArray(Content):
             self._length,
             self._identifier,
             ak._v2._util.merge_parameters(self._parameters, parameters),
+            self._nplike,
         )
 
     def index_to_field(self, index):
@@ -205,7 +240,7 @@ class RecordArray(Content):
 
     def content(self, index_or_field):
         out = self.Form.content(self, index_or_field)
-        if len(out) == self._length:
+        if out.length == self._length:
             return out
         else:
             return out[: self._length]
@@ -214,14 +249,18 @@ class RecordArray(Content):
         return self._getitem_range(slice(0, 0))
 
     def _getitem_at(self, where):
-        if where < 0:
-            where += len(self)
-        if not (0 <= where < len(self)) and self.nplike.known_shape:
+        if self._nplike.known_data and where < 0:
+            where += self.length
+
+        if where < 0 or where >= self.length:
             raise NestedIndexError(self, where)
         return Record(self, where)
 
     def _getitem_range(self, where):
-        start, stop, step = where.indices(len(self))
+        if not self._nplike.known_shape:
+            return self
+
+        start, stop, step = where.indices(self.length)
         assert step == 1
         if len(self._contents) == 0:
             start = min(max(start, 0), self._length)
@@ -234,6 +273,7 @@ class RecordArray(Content):
                 stop - start,
                 self._range_identifier(start, stop),
                 self._parameters,
+                self._nplike,
             )
         else:
             nextslice = slice(start, stop)
@@ -243,6 +283,7 @@ class RecordArray(Content):
                 stop - start,
                 self._range_identifier(start, stop),
                 self._parameters,
+                self._nplike,
             )
 
     def _getitem_field(self, where, only_fields=()):
@@ -279,24 +320,26 @@ class RecordArray(Content):
         return RecordArray(
             contents,
             fields,
+            self._length,
             self._fields_identifier(where),
             None,
+            self._nplike,
         )
 
     def _carry(self, carry, allow_lazy, exception):
         assert isinstance(carry, ak._v2.index.Index)
 
         if allow_lazy:
-            nplike, where = carry.nplike, carry.data
+            where = carry.data
 
             if allow_lazy != "copied":
                 where = where.copy()
 
             negative = where < 0
-            if nplike.any(negative, prefer=False):
+            if self._nplike.any(negative, prefer=False):
                 where[negative] += self._length
 
-            if nplike.any(where >= self._length, prefer=False):
+            if self._nplike.any(where >= self._length, prefer=False):
                 if issubclass(exception, NestedIndexError):
                     raise exception(self, where)
                 else:
@@ -308,6 +351,7 @@ class RecordArray(Content):
                 self,
                 self._carry_identifier(carry, exception),
                 None,
+                self._nplike,
             )
 
         else:
@@ -317,11 +361,11 @@ class RecordArray(Content):
             ]
 
             # if issubclass(carry.dtype.type, np.integer):
-            #     length = len(carry)
+            #     length = carry.length
             # else:
-            #     length = len(self)
+            #     length = self.length
             assert issubclass(carry.dtype.type, np.integer)
-            length = len(carry)
+            length = carry.length
 
             return RecordArray(
                 contents,
@@ -329,6 +373,7 @@ class RecordArray(Content):
                 length,
                 self._carry_identifier(carry, exception),
                 self._parameters,
+                self._nplike,
             )
 
     def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
@@ -339,7 +384,9 @@ class RecordArray(Content):
                     slicestarts, slicestops, slicecontent, tail
                 )
             )
-        return RecordArray(contents, self._fields, self._length)
+        return RecordArray(
+            contents, self._fields, self._length, None, None, self._nplike
+        )
 
     def _getitem_next(self, head, tail, advanced):
         if head == ():
@@ -376,20 +423,22 @@ class RecordArray(Content):
             ):
                 parameters = self._parameters
             next = RecordArray(
-                contents, self._fields, None, self._identifier, parameters
+                contents, self._fields, None, self._identifier, parameters, self._nplike
             )
             return next._getitem_next(nexthead, nexttail, advanced)
 
     def num(self, axis, depth=0):
         posaxis = self.axis_wrap_if_negative(axis)
         if posaxis == depth:
-            single = ak._v2.index.Index64.empty(1, self.nplike)
-            single[0] = len(self)
-            singleton = ak._v2.contents.numpyarray.NumpyArray(single)
+            single = ak._v2.index.Index64.empty(1, self._nplike)
+            single[0] = self.length
+            singleton = ak._v2.contents.numpyarray.NumpyArray(
+                single, None, None, self._nplike
+            )
             contents = [singleton] * len(self._contents)
 
             record = ak._v2.contents.recordarray.RecordArray(
-                contents, self._fields, 1, None, self._parameters
+                contents, self._fields, 1, None, self._parameters, self._nplike
             )
             return record[0]
         else:
@@ -397,7 +446,12 @@ class RecordArray(Content):
             for content in self._contents:
                 contents.append(content.num(posaxis, depth))
             return ak._v2.contents.recordarray.RecordArray(
-                contents, self._fields, self._length, None, self._parameters
+                contents,
+                self._fields,
+                self._length,
+                None,
+                self._parameters,
+                self._nplike,
             )
 
     def _offsets_and_flattened(self, axis, depth):
@@ -413,18 +467,23 @@ class RecordArray(Content):
         else:
             contents = []
             for content in self._contents:
-                trimmed = content._getitem_range(slice(0, len(self)))
+                trimmed = content._getitem_range(slice(0, self.length))
                 offsets, flattened = trimmed._offsets_and_flattened(posaxis, depth)
-                if len(offsets) != 0:
+                if self._nplike.known_shape and offsets.length != 0:
                     raise AssertionError(
                         "RecordArray content with axis > depth + 1 returned a non-empty offsets from offsets_and_flattened"
                     )
                 contents.append(flattened)
-            offsets = ak._v2.index.Index64.zeros(1, self.nplike, dtype=np.int64)
+            offsets = ak._v2.index.Index64.zeros(1, self._nplike, dtype=np.int64)
             return (
                 offsets,
                 RecordArray(
-                    contents, self._fields, self._length, None, self._parameters
+                    contents,
+                    self._fields,
+                    self._length,
+                    None,
+                    {},
+                    self._nplike,
                 ),
             )
 
@@ -453,25 +512,29 @@ class RecordArray(Content):
         ):
             return self.mergeable(other.content, mergebool)
 
-        if isinstance(other, ak._v2.contents.recordarray.RecordArray):
+        if isinstance(other, RecordArray):
             if self.is_tuple and other.is_tuple:
-                if len(self.contents) == len(other.contents):
-                    for i in range(len(self.contents)):
-                        if not self.contents[i].mergeable(other.contents[i], mergebool):
+                if len(self._contents) == len(other._contents):
+                    for self_cont, other_cont in zip(self._contents, other._contents):
+                        if not self_cont.mergeable(other_cont, mergebool):
                             return False
+
                     return True
 
             elif not self.is_tuple and not other.is_tuple:
-                self_fields = self.fields.copy()
-                other_fields = other.fields.copy()
-                self_fields.sort()
-                other_fields.sort()
-                if self_fields == other_fields:
-                    for field in self_fields:
-                        if not self[field].mergeable(other[field], mergebool):
-                            return False
-                    return True
-            return False
+                if set(self._fields) != set(other._fields):
+                    return False
+
+                for i, field in enumerate(self._fields):
+                    x = self._contents[i]
+                    y = other._contents[other.field_to_index(field)]
+                    if not x.mergeable(y, mergebool):
+                        return False
+                return True
+
+            else:
+                return False
+
         else:
             return False
 
@@ -481,18 +544,18 @@ class RecordArray(Content):
 
         head, tail = self._merging_strategy(others)
 
-        parameters = {}
+        parameters = self._parameters
         headless = head[1:]
 
         for_each_field = []
         for field in self.contents:
-            trimmed = field[0 : len(self)]
+            trimmed = field[0 : self.length]
             for_each_field.append([field])
 
         if self.is_tuple:
             for array in headless:
                 parameters = ak._v2._util.merge_parameters(
-                    self._parameters, array._parameters
+                    self._parameters, array._parameters, True
                 )
 
                 if isinstance(array, ak._v2.contents.recordarray.RecordArray):
@@ -500,7 +563,7 @@ class RecordArray(Content):
                         if len(self.contents) == len(array.contents):
                             for i in range(len(self.contents)):
                                 field = array[self.index_to_field(i)]
-                                for_each_field[i].append(field[0 : len(array)])
+                                for_each_field[i].append(field[0 : array.length])
                         else:
                             raise ValueError(
                                 "cannot merge tuples with different numbers of fields"
@@ -531,7 +594,7 @@ class RecordArray(Content):
                             for i in range(len(self.contents)):
                                 field = array[self.index_to_field(i)]
 
-                                trimmed = field[0 : len(array)]
+                                trimmed = field[0 : array.length]
                                 for_each_field[i].append(trimmed)
                         else:
                             raise AssertionError(
@@ -558,15 +621,17 @@ class RecordArray(Content):
 
             nextcontents.append(merged)
 
-            if minlength is None or len(merged) < minlength:
-                minlength = len(merged)
+            if minlength is None or merged.length < minlength:
+                minlength = merged.length
 
         if minlength is None:
-            minlength = len(self)
+            minlength = self.length
             for x in others:
-                minlength += len(x)
+                minlength += x.length
 
-        next = RecordArray(nextcontents, self._fields, minlength, None, parameters)
+        next = RecordArray(
+            nextcontents, self._fields, minlength, None, parameters, self._nplike
+        )
 
         if len(tail) == 0:
             return next
@@ -586,7 +651,12 @@ class RecordArray(Content):
         for content in self._contents:
             contents.append(content.fillna(value))
         return RecordArray(
-            contents, self._fields, self._length, self._identifier, self._parameters
+            contents,
+            self._fields,
+            self._length,
+            self._identifier,
+            self._parameters,
+            self._nplike,
         )
 
     def _localindex(self, axis, depth):
@@ -598,7 +668,12 @@ class RecordArray(Content):
             for content in self._contents:
                 contents.append(content._localindex(posaxis, depth))
             return RecordArray(
-                contents, self._fields, len(self), self._identifier, self._parameters
+                contents,
+                self._fields,
+                self.length,
+                self._identifier,
+                self._parameters,
+                self._nplike,
             )
 
     def numbers_to_type(self, name):
@@ -606,7 +681,12 @@ class RecordArray(Content):
         for x in self._contents:
             contents.append(x.numbers_to_type(name))
         return ak._v2.contents.recordarray.RecordArray(
-            contents, self._fields, self._length, self._identifier, self._parameters
+            contents,
+            self._fields,
+            self._length,
+            self._identifier,
+            self._parameters,
+            self._nplike,
         )
 
     def _is_unique(self, negaxis, starts, parents, outlength):
@@ -631,7 +711,9 @@ class RecordArray(Content):
         order,
     ):
         if self._fields is None or len(self._fields) == 0:
-            return ak._v2.contents.NumpyArray(self.nplike.empty(0, np.int64))
+            return ak._v2.contents.NumpyArray(
+                self._nplike.empty(0, np.int64), None, None, self._nplike
+            )
         else:
             raise NotImplementedError
 
@@ -639,7 +721,9 @@ class RecordArray(Content):
         self, negaxis, starts, parents, outlength, ascending, stable, kind, order
     ):
         if self._fields is None or len(self._fields) == 0:
-            return ak._v2.contents.NumpyArray(self.nplike.instance().empty(0, np.int64))
+            return ak._v2.contents.NumpyArray(
+                self._nplike.instance().empty(0, np.int64), None, None, self._nplike
+            )
 
         contents = []
         for content in self._contents:
@@ -655,7 +739,9 @@ class RecordArray(Content):
                     order,
                 )
             )
-        return RecordArray(contents, self._fields, len(self), None, self._parameters)
+        return RecordArray(
+            contents, self._fields, self._length, None, self._parameters, self._nplike
+        )
 
     def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
         posaxis = self.axis_wrap_if_negative(axis)
@@ -670,7 +756,12 @@ class RecordArray(Content):
                     )
                 )
             return ak._v2.contents.recordarray.RecordArray(
-                contents, recordlookup, len(self), self._identifier, self._parameters
+                contents,
+                recordlookup,
+                self.length,
+                self._identifier,
+                self._parameters,
+                self._nplike,
             )
 
     def _reduce_next(
@@ -705,16 +796,15 @@ class RecordArray(Content):
             outlength,
             None,
             None,
+            self._nplike,
         )
 
     def _validityerror(self, path):
-        for i in range(len(self.contents)):
-            if len(self.contents[i]) < len(self):
-                return 'at {0} ("{1}"): len(field({2})) < len(recordarray)'.format(
-                    path, type(self), i
-                )
-        for i in range(len(self.contents)):
-            sub = self.contents[i].validityerror(path + ".field({0})".format(i))
+        for i, cont in enumerate(self.contents):
+            if cont.length < self.length:
+                return f'at {path} ("{type(self)}"): len(field({i})) < len(recordarray)'
+        for i, cont in enumerate(self.contents):
+            sub = cont.validityerror(f"{path}.field({i})")
             if sub != "":
                 return sub
         return ""
@@ -739,21 +829,24 @@ class RecordArray(Content):
                 return ak._v2.contents.recordarray.RecordArray(
                     contents,
                     self._fields,
-                    len(self),
-                    identifier=self._identifier,
-                    parameters=self._parameters,
+                    self._length,
+                    self._identifier,
+                    self._parameters,
+                    self._nplike,
                 )
             else:
                 return ak._v2.contents.recordarray.RecordArray(
                     contents,
                     self._fields,
-                    identifier=self._identifier,
-                    parameters=self._parameters,
+                    self._length,
+                    self._identifier,
+                    self._parameters,
+                    self._nplike,
                 )
 
     def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
         values = [
-            (x if len(x) == length else x[:length])._to_arrow(
+            (x if x.length == length else x[:length])._to_arrow(
                 pyarrow, mask_node, validbytes, length, options
             )
             for x in self._contents
@@ -779,27 +872,27 @@ class RecordArray(Content):
 
     def _to_numpy(self, allow_missing):
         if self.fields is None:
-            return self.nplike.empty(len(self), dtype=[])
+            return self._nplike.empty(self.length, dtype=[])
         contents = [x._to_numpy(allow_missing) for x in self._contents]
         if any(len(x.shape) != 1 for x in contents):
-            raise ValueError("cannot convert {0} into np.ndarray".format(self))
-        out = self.nplike.empty(
-            len(contents[0]),
+            raise ValueError(f"cannot convert {self} into np.ndarray")
+        out = self._nplike.empty(
+            contents[0].shape[0],
             dtype=[(str(n), x.dtype) for n, x in zip(self.fields, contents)],
         )
         mask = None
         for n, x in zip(self.fields, contents):
-            if isinstance(x, self.nplike.ma.MaskedArray):
+            if isinstance(x, self._nplike.ma.MaskedArray):
                 if mask is None:
-                    mask = self.nplike.ma.zeros(
-                        len(self), [(n, np.bool_) for n in self.fields]
+                    mask = self._nplike.ma.zeros(
+                        self.length, [(n, np.bool_) for n in self.fields]
                     )
                 if x.mask is not None:
                     mask[n] |= x.mask
             out[n] = x
 
         if mask is not None:
-            out = self.nplike.ma.MaskedArray(out, mask)
+            out = self._nplike.ma.MaskedArray(out, mask)
 
         return out
 
@@ -814,7 +907,7 @@ class RecordArray(Content):
             if options["function_name"] is not None:
                 in_function = " in " + options["function_name"]
             raise TypeError(
-                "cannot combine record fields{0} unless flatten_records=True".format(
+                "cannot combine record fields{} unless flatten_records=True".format(
                     in_function
                 )
             )
@@ -840,6 +933,7 @@ class RecordArray(Content):
                     self._length,
                     self._identifier,
                     self._parameters if options["keep_parameters"] else None,
+                    self._nplike,
                 )
 
         else:
@@ -873,13 +967,14 @@ class RecordArray(Content):
     def packed(self):
         return RecordArray(
             [
-                x.packed() if len(x) == self._length else x[: self._length].packed()
+                x.packed() if x.length == self._length else x[: self._length].packed()
                 for x in self._contents
             ],
             self._fields,
             self._length,
             self._identifier,
             self._parameters,
+            self._nplike,
         )
 
     def _to_list(self, behavior):
@@ -911,3 +1006,72 @@ class RecordArray(Content):
             for i in range(length):
                 out[i] = dict(zip(fields, [x[i] for x in contents]))
             return out
+
+    def _to_nplike(self, nplike):
+        contents = [content._to_nplike(nplike) for content in self._contents]
+        return RecordArray(
+            contents,
+            self._fields,
+            length=self._length,
+            identifier=self._identifier,
+            parameters=self._parameters,
+            nplike=nplike,
+        )
+
+    def _to_json(
+        self,
+        nan_string,
+        infinity_string,
+        minus_infinity_string,
+        complex_real_string,
+        complex_imag_string,
+    ):
+        out = self._to_json_custom()
+        if out is not None:
+            return out
+
+        cls = ak._v2._util.recordclass(self, None)
+        if cls is not ak._v2.highlevel.Record:
+            length = self._length
+            out = [None] * length
+            for i in range(length):
+                out[i] = cls(self[i])
+            return out
+
+        if self.is_tuple:
+            contents = [
+                x._to_json(
+                    nan_string,
+                    infinity_string,
+                    minus_infinity_string,
+                    complex_real_string,
+                    complex_imag_string,
+                )
+                for x in self._contents
+            ]
+            length = self._length
+            out = [None] * length
+            fields = []
+            for i in range(length):
+                fields.append(str(i))
+            for i in range(length):
+                out[i] = dict(zip(fields, [x[i] for x in contents]))
+            return out
+
+        else:
+            fields = self._fields
+            contents = [
+                x._to_json(
+                    nan_string,
+                    infinity_string,
+                    minus_infinity_string,
+                    complex_real_string,
+                    complex_imag_string,
+                )
+                for x in self._contents
+            ]
+            length = self._length
+            out = [None] * length
+            for i in range(length):
+                out[i] = dict(zip(fields, [x[i] for x in contents]))
+            return out
